/*
 * Ethernet Driver Template - PCI
 * CopyRight (c) 1999 Be Inc., All Rights Reserved.
 * This file may be used under the terms of the Be Sample Code License.
 */
/*
 * Copyright (c) 1997, 1998, 1999
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD: src/sys/pci/if_sis.c,v 1.58 2002/11/14 23:49:09 sam Exp $
 */
 /*
 * Now supports NICs based off the DP83815 chipset 
 * Modified: 12/10/02
 * By: Antonio Carpio
 * E-mail: BolivianTONE@nc.rr.com
 *
 * Notes:
 *	-From my computer the throughput eventually trails and stalls.  No
 *	 clue yet as to why.
 *	-The above copyright pertains to the sis functions at the end of the
 *	 source since I didn't know how to access the eeprom.
 * To-dos:
 *	-Fix KDL visit when you restart the NetServer under certain conditions
 *	-Read configuration in from config file
 *	-Implement readv, writev
 *	-implement multicast hashing
 *	-things not thought of right now
 *
 *	Please send any comments or fixes back to me as this is a running project for me.
 *  I have a few other cards I am also going to be working:
 *		DP83820 based cards
 *		SMC 1244TX
 *		Linksys Etherlink of some sort
 */
 
 
#include <OS.h>
#include <KernelExport.h>
#include <Drivers.h>
#include <stdarg.h>
#include <PCI.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <SupportDefs.h>
#include <ByteOrder.h>
#include <kernel/image.h>

#include "../ether_driver.h"
#include "if_dp83815_regs.h"


/* PCI vendor and d ID's */
#define VENDOR_ID        0x100B
#define DEVICE_ID        0x0020

/* globals for driver instances */ 

#define kDevName "DP83815"
#define kDevDir "net/" kDevName "/"
#define DEVNAME_LENGTH		64			
#define MAX_CARDS 			4			/* maximum number of driver instances	*/
#define MAX_DESC			128			/* eventually going into config file	*/
#define	BUFFER_SIZE			2048		/* make it easy to do math and whatnot	*/
#define MAX_PACKET_SIZE		1518		/* can be 2046 if RXCFG has ALP set		*/
#define	NONBLOCK_WAIT		1000



#define read32(address) 				*((volatile uint32 *)(d->reg_base + address));
#define write32(address, data) 			(*((volatile uint32 *)(d->reg_base + address)) = data); __eieio()



static pci_module_info		*gPCIModInfo;
static char 				*gDevNameList[MAX_CARDS+1];
static pci_info 			*gDevList[MAX_CARDS+1];
static uint32 				gOpenMask = 0;



const char** publish_devices(void )	{ return (const char **)gDevNameList; }


/* Driver Entry Points */
__declspec(dllexport) status_t init_hardware(void );
__declspec(dllexport) status_t init_driver(void );
__declspec(dllexport) void uninit_driver(void );
__declspec(dllexport) const char** publish_devices(void );
__declspec(dllexport) device_hooks *find_device(const char *name );





typedef struct descriptor {
	uint32				link;			/*	Physical address of next descriptor	*/
	volatile uint32		cmd;			/*	info to/from NIC					*/
	volatile uint32		ptr;			/*	Physical address of the buffer		*/
	struct descriptor	*virt_next;		/*	virtual address of next descriptor	*/
	void				*virt_buff;		/*	virtual address of the buffer		*/
} descriptor_t;



typedef struct desc_ring {
	descriptor_t	*Curr;				/*	Current descriptor to be accessed	*/
	sem_id			Sem;				/*	Queue for outstanding reads/writes	*/
	spinlock		Lock;				/*	Spinlock used to grab a descriptor	*/
	descriptor_t	*CurrInt;			/*	Interrupt Hook's Current Descriptor	*/
} desc_ring_t;



/* per driver intance globals */
typedef struct {
	ether_address_t mac_address;		/* my ethernet address */
	int32			devID; 				/* d identifier: 0-n */
	pci_info		*pciInfo;
	uint32			reg_base; 			/* Base address of hostRegs */
	area_id			reg_area,
					mem_area;						

	int32			blockFlag;			/* for blocking or nonblocking reads */

	uint32			media;				/* speed duplex media settings */

	stats_t			stats;
	desc_ring_t		Rx, Tx;
	
} dev_info_t;




/* prototypes */
static status_t open_hook(const char *name, uint32 flags, void **_cookie);
static status_t close_hook(void *_d);
static status_t free_hook(void *_d);
static status_t control_hook(void * cookie,uint32 msg,void *buf,size_t len);
static status_t read_hook(void *_d,off_t pos, void *buf,size_t *len);
static status_t write_hook(void *_d, off_t pos, const void *buf, size_t *len);
static int32	interrupt_hook(void *_d);                	   /* interrupt handler */


static int32    get_pci_list(pci_info *info[], int32 maxEntries);  /* Get pci_info for each d */
static status_t free_pci_list(pci_info *info[]);                   /* Free storage used by pci_info list */
static status_t enable_addressing(dev_info_t *d);             /* enable pci io address space for d */
static status_t init_ring_buffers(dev_info_t *d);             /* allocate and initialize frame buffer rings */
static void     get_mac_address(dev_info_t *d);               /* get ethernet address */
static status_t setpromisc(dev_info_t * d);                   /* set hardware to receive all packets */
static status_t domulti(dev_info_t *d, uint8 *addr);		   /* add multicast address to hardware filter list */
static status_t allocate_resources(dev_info_t *d);     	   /* allocate semaphores & spinlocks */
static void     free_resources(dev_info_t *d);                /* deallocate semaphores & spinlocks */


static uint16 sis_eeprom_getword(dev_info_t	*d, int addr);





static device_hooks gDeviceHooks =  {
	open_hook, 			/* -> open entry point */
	close_hook,         /* -> close entry point */
	free_hook,          /* -> free entry point */
	control_hook,       /* -> control entry point */
	read_hook,          /* -> read entry point */
	write_hook,         /* -> write entry point */
	NULL,               /* -> select entry point */
	NULL,                /* -> deselect entry point */
	NULL,               /* -> readv */
	NULL                /* -> writev */
};


uint32 pages_needed(uint32 mem_size)
{
	uint32 pages = mem_size/B_PAGE_SIZE;
	if( pages%B_PAGE_SIZE != 0 ) pages += 1;
	return pages;
}


void print_desc(struct descriptor* d)
{
	dprintf("link: %x\ncmd: %x\nptr: %x\nvirt_next: %x\nvirt_buff: %x\n",	d->link, d->cmd, d->ptr, d->virt_next, d->virt_buff);
}


//  Driver Entry Points  //
status_t init_hardware(void )
{
	return B_OK;
}



status_t init_driver()
{
	status_t 		status;
	int32			entries;
	char			devName[64];
	int32			i;
	
	if((status = get_module( B_PCI_MODULE_NAME, (module_info **)&gPCIModInfo )) != B_OK) {
		dprintf(kDevName " Get module failed! %s\n", strerror(status ));
		return status;
	}
		
	//  Find Lan cards
	if ((entries = get_pci_list(gDevList, MAX_CARDS )) == 0) {
		dprintf(kDevName "No devices supported by DP83815 found\n");
		free_pci_list(gDevList);
		put_module(B_PCI_MODULE_NAME );
		return B_ERROR;
	}
	
	//  Create device name list
	for (i=0; i<entries; i++ ) {
		sprintf(devName, "%s%ld", kDevDir, i );
		gDevNameList[i] = (char *)malloc(strlen(devName)+1);
		strcpy(gDevNameList[i], devName);
	}
	gDevNameList[i] = NULL;
	
	return B_OK;
}

void uninit_driver(void)
{
	int32 	i;
	void 	*item;

	/*Free d name list*/
	for (i=0; (item=gDevNameList[i]); i++)
		free(item);
	
	/*Free d list*/
	free_pci_list(gDevList);
	put_module(B_PCI_MODULE_NAME);
}


device_hooks *find_device(const char *name)
{
	int32 	i;
	char 	*item;

	for( i=0; (item=gDevNameList[i]); i++ )		//	For each published device name
		if( strcmp(name, item) == 0 )			//	  if that is the wanted one
			return &gDeviceHooks;					//	    return this driver's d hooks
												//	Otherwise
	return NULL;								//	  d not found
}




static status_t open_hook(const char *name, uint32 flags, void **cookie)
{
	int32			devID;
	int32			mask;
	status_t		status;
	char 			*devName;
	dev_info_t 		*d;
	

	/*	Find d name*/
	for (devID=0; (devName=gDevNameList[devID]); devID++) {
		if (strcmp(name, devName) == 0)
			break;
	}
	if (!devName)
		return EINVAL;
	
	/* Check if the d is busy and set in-use flag if not */
	mask = 1 << devID;
	if (atomic_or(&gOpenMask, mask) &mask) {
		return B_BUSY;
	}

	/*	Allocate storage for the cookie*/
	if (!(*cookie = d = (dev_info_t *)malloc(sizeof(dev_info_t)))) {
		status = B_NO_MEMORY;
		goto err0;
	}
	
	memset(d, 0, sizeof(dev_info_t));
	
	/* Setup the cookie */
	d->pciInfo = gDevList[devID];
	d->devID = devID;


	if ( enable_addressing(d) != B_OK)
		goto err1;

	if (allocate_resources(d) != B_OK)
		goto err1;
	
	get_mac_address(d);
	install_io_interrupt_handler( d->pciInfo->u.h0.interrupt_line,
								  interrupt_hook,
								  *cookie,
								  0 );
	{
		uint32 temp = read32(REG_SRR);
		dprintf("SRR: %x\n", temp);
	}
								  
	write32(REG_CR, CR_RXR|CR_TXR);			/*	Reset Tx & Rx		*/
	
	if ( init_ring_buffers(d) != B_OK )		/*	Init ring buffers	*/
		goto err1;
	
	write32(REG_RFCR, RFCR_RFEN|RFCR_AAB|RFCR_AAM|RFCR_AAU);
	
	write32(REG_RXCFG, RXCFG_ATP|RXCFG_DRTH(31));					/*	Set the drth		*/
	
	write32(REG_TXCFG, TXCFG_CSI|
						TXCFG_HBI|
						TXCFG_ATP|
						TXCFG_MXDMA_256|
						TXCFG_FLTH(16)|
						TXCFG_DRTH(16) );
	
	write32(REG_IMR, ISR_RXIDLE | ISR_TXOK | ISR_RXOK );
	
	write32(REG_CR, CR_RXE);			/*	Enable Rx				*/
	write32(REG_IER, 1);				/*	Enable interrupts		*/
	
	return B_NO_ERROR;
	
	err1:
		free_resources(d);
		free(d);	
	err0:
		atomic_and(&gOpenMask, ~mask);
		return status;
}



static status_t close_hook(void *_d) {
	dev_info_t 	*d =  (dev_info_t *) _d;
	int			irq = d->pciInfo->u.h0.interrupt_line;
	
	write32(REG_IER, 0);				/*	Disable interrupts		*/
	write32(REG_CR, CR_RXD|CR_TXD);		/*	Disable Rx & Tx			*/
	
	while ( d->Tx.Lock );				/*	wait for any current writes to finish	*/
	while ( d->Rx.Lock );				/*	wait for any current reads to finish	*/
	
	remove_io_interrupt_handler(irq,
								interrupt_hook,
								_d);
	
	free_resources(d);					/*	unblock waiting threads		*/
	
	atomic_and(&gOpenMask, ~(1 << d->devID));		/*	device is available again	*/
	
	return B_NO_ERROR;
}



static status_t free_hook(void * cookie) {
	dev_info_t *d = (dev_info_t *) cookie;
	
	delete_area(d->reg_area);
	delete_area(d->mem_area);
	
	free(d);
	
	return 0;
}



/* Standard driver control function */
static status_t control_hook(void * cookie,uint32 msg,void *buf,size_t len)
{
	dev_info_t *d = (dev_info_t *) cookie;

	switch (msg) {
		case ETHER_GETADDR:
		{
			uint8 i;
			for (i=0; i<6; i++)
				((char*)buf)[i] = d->mac_address.ebyte[i];
			return B_NO_ERROR;
		}

		case ETHER_INIT:
			return B_NO_ERROR;
			
		case ETHER_GETFRAMESIZE:
			*(uint32 *)buf = MAX_PACKET_SIZE;
			return B_NO_ERROR;
			
		case ETHER_ADDMULTI:
			return domulti(d, (unsigned char *)buf);
		
		case ETHER_SETPROMISC:
			write32(REG_RFCR, RFCR_RFEN|RFCR_AAB|RFCR_AAU|RFCR_AAM);
			return B_NO_ERROR;
		
		case ETHER_NONBLOCK:
			if (*((int32 *)buf))
				d->blockFlag = B_RELATIVE_TIMEOUT;
			else
				d->blockFlag = 0;
			return B_NO_ERROR;

		case ETHER_GETSTATS:
			memcpy(buf, &d->stats, sizeof(stats_t));
			return B_NO_ERROR;
			
		default:
			dprintf("Hmmmm, unknown control msg\n");
			return B_ERROR;
	}
}



/* The read() system call - upper layer interface to the ethernet driver */
static status_t  read_hook(void *_d,off_t pos, void *buf, size_t *len)
{
	dev_info_t		*d = (dev_info_t *) _d;
	status_t		s;
	size_t			length = 0;
	descriptor_t*	desc;
	
	if ( B_NO_ERROR != (s = acquire_sem_etc(d->Rx.Sem,1,B_CAN_INTERRUPT|d->blockFlag,NONBLOCK_WAIT)) ) {
		*len = 0;
		return s;
	}
	
	/*	Grab the first descriptor and stop blocking	*/
	{
		cpu_status former;
		former = disable_interrupts();
		acquire_spinlock(&d->Rx.Lock);
			desc = d->Rx.Curr;
			d->Rx.Curr = desc->virt_next;
		release_spinlock(&d->Rx.Lock);
		restore_interrupts(former);
	}
	
	length= DESC_LENGTH&desc->cmd;
	
	if( desc->cmd & (DESC_RXA|DESC_RXO|DESC_LONG|DESC_RUNT|DESC_ISE|DESC_CRCE|DESC_FAE|DESC_LBP|DESC_COL) )
		dprintf("desc cmd: %x\n", desc->cmd );
	
	if( length < 64 ) {
		*len = 0;
		return B_ERROR;
	}
	
	if( *len < length )
		length = *len;
		
	memcpy(buf, desc->virt_buff,length);
	desc->cmd = DESC_LENGTH&MAX_PACKET_SIZE;
	*len = length;

	atomic_add(&d->stats.rx_att, 1);

	if( length == 0 )
		return B_ERROR;
	return B_OK;
}


/*
 * The write() system call - upper layer interface to the ethernet driver
 */
static status_t write_hook(void *_d, off_t pos, const void *buf, size_t *len)
{
	dev_info_t*		d = (dev_info_t *) _d;
	descriptor_t*	desc;
	status_t		s;	

	if ( B_NO_ERROR != (s = acquire_sem_etc(d->Tx.Sem,1,B_CAN_INTERRUPT|d->blockFlag,NONBLOCK_WAIT)) ) {
		*len = 0;
		return s;
	}

	/*	Grab the first descriptor and stop blocking	*/
	{
		cpu_status former;
		former = disable_interrupts();
		acquire_spinlock(&d->Tx.Lock);
			desc = d->Tx.Curr;
			d->Tx.Curr = desc->virt_next;
		release_spinlock(&d->Tx.Lock);
		restore_interrupts(former);
	}


	if ( *len > MAX_PACKET_SIZE ) {			/* if needed						*/
		dprintf("Had to truncate the packet from %d to %d\n", *len, MAX_PACKET_SIZE);
		*len = MAX_PACKET_SIZE;				/*   truncate the packet			*/
	}


	while( desc->cmd&DESC_OWN )	{			/* make sure a buffer is available	*/
		dprintf("spinning in the write hook\n");
		spin(1000);						/* wait a while if not				*/
	}

	memcpy(desc->virt_buff, buf, *len);		/* now copy the data				*/
	desc->cmd = DESC_OWN|*len;				/* update the cmd bits				*/

	write32(REG_CR, CR_TXE);				/* tell the card to start tx		*/
	atomic_add(&d->stats.tx_att, 1);
	return B_OK;
}


/* interrupt service routine */
static int32
interrupt_hook(void *_d)
{
	dev_info_t *d = (dev_info_t *) _d;
	uint32		isr;
	isr = read32(REG_ISR);

	if ( isr == 0 )	return B_UNHANDLED_INTERRUPT;

	if ( isr & ISR_RXOK ) {
		int				num_packets = 0; 		
		descriptor_t	*curr = d->Rx.CurrInt;	

		while( curr->cmd & DESC_OWN ) {				
			curr = curr->virt_next;					
			num_packets++;							
		}
				
		d->Rx.CurrInt = curr;
		d->stats.rx_ok += num_packets;
		if( num_packets > 1 )
			dprintf("received %d descriptors\n", num_packets);
		if( num_packets )
			release_sem_etc(d->Rx.Sem, num_packets, B_DO_NOT_RESCHEDULE);
	}
	
	if( isr & ISR_TXOK ) {
		d->stats.tx_ok++;
		release_sem_etc(d->Tx.Sem, 1, B_DO_NOT_RESCHEDULE);
	}
	
	if( isr & ISR_RXIDLE )
		dprintf("DAMMIT!  RX IS IDLE!\n");
		
	if( isr & ~(ISR_TXIDLE|ISR_TXOK|ISR_RXOK|ISR_RXIDLE|ISR_RXEARLY) )
		dprintf("ISR: %x\n", isr);
	
	return B_INVOKE_SCHEDULER;
}



static int32 get_pci_list(pci_info *info[], int32 maxEntries)
{
	status_t status;
	int32 i, entries;
	pci_info	*item;
	
	item = (pci_info *)malloc(sizeof(pci_info));
	
	for (i=0, entries=0; entries<maxEntries; i++) {
		if ((status = gPCIModInfo->get_nth_pci_info(i, item)) != B_OK)
			break;
		if ( 0 || (item->vendor_id == VENDOR_ID)&& (item->device_id == DEVICE_ID)) {
			/* check if the d really has an IRQ */
			if ((item->u.h0.interrupt_line == 0) || (item->u.h0.interrupt_line == 0xFF)) {
				dprintf(kDevName " supported device found with invalid IRQ - check IRQ assignement\n");
				continue;
			}
			dprintf(kDevName " supported device found at IRQ %x\n", item->u.h0.interrupt_line);
			info[entries++] = item;
			item = (pci_info *)malloc(sizeof(pci_info));
		}
	}
	info[entries] = NULL;
	free(item);
	return entries;
}

static status_t free_pci_list(pci_info *info[])
{
	int32 		i;
	pci_info	*item;
	
	for (i=0; (item=info[i]); i++)
		free(item);
	return B_OK;
}


static status_t enable_addressing(dev_info_t *d)
{
	//  Turn on Memory Address Decode, and Bus Mastering - this may or may not be necessary for your hardware */
	gPCIModInfo->write_pci_config(d->pciInfo->bus, d->pciInfo->device, 
		d->pciInfo->function, PCI_command, 2,
		PCI_command_memory | PCI_command_master |
		gPCIModInfo->read_pci_config( d->pciInfo->bus, d->pciInfo->device, 
		d->pciInfo->function, PCI_command, 2));


	//  Enable memory mapping
	{
		int32		base, size, offset, i;
		int32		*ptr;
		const		MemoryReg = 1;
	
		base = d->pciInfo->u.h0.base_registers[MemoryReg];			//	
		size = d->pciInfo->u.h0.base_register_sizes[MemoryReg];	//	
	
		//  Round down to nearest page boundary
		base = base & ~(B_PAGE_SIZE-1);
	
		//  Adjust the size
		offset = d->pciInfo->u.h0.base_registers[MemoryReg] - base;
		size += offset;
		size = (size +(B_PAGE_SIZE-1)) & ~(B_PAGE_SIZE-1);

		d->reg_area = map_physical_memory(kDevName " Regs",
											(void *)base,
											size,
											B_ANY_KERNEL_ADDRESS,
											B_READ_AREA | B_WRITE_AREA | B_MTR_UC,
											(void**)&d->reg_base);
		if( d->reg_area < 0 )
			return B_ERROR;

		d->reg_base = d->reg_base + offset;
	}
	
	return B_OK;
}




static status_t init_ring_buffers(dev_info_t *d)
{
	uint32			i;
	area_info		info;
	physical_entry	map[2];
	uint32 pages;
	
	descriptor_t	*RxDescRing = NULL;
	descriptor_t	*TxDescRing = NULL;

	descriptor_t	*desc_base_virt_addr;
	uint32			desc_base_phys_addr;
	
	void			*buff_base_virt_addr;
	uint32			buff_base_phys_addr;


	d->mem_area = 0;		

#define NUM_BUFFS	2*MAX_DESC
	pages = pages_needed(2*MAX_DESC*sizeof(descriptor_t) + NUM_BUFFS*BUFFER_SIZE);

	d->mem_area = create_area(kDevName" desc buffer",
							  (void**)&RxDescRing,
							  B_ANY_KERNEL_ADDRESS,
							  pages*B_PAGE_SIZE,
							  B_FULL_LOCK|B_CONTIGUOUS,
							  B_READ_AREA|B_WRITE_AREA);
	if( d->mem_area < 0 )
		return -1;


	get_area_info(d->mem_area, &info);
	get_memory_map(info.address, info.size, map, 4);
	
	desc_base_phys_addr = (int)map[0].address + NUM_BUFFS*BUFFER_SIZE;
	desc_base_virt_addr = (info.address + NUM_BUFFS*BUFFER_SIZE);
	
	buff_base_phys_addr = (int)map[0].address;
	buff_base_virt_addr = info.address;
	

	RxDescRing = desc_base_virt_addr;
	for( i = 0; i < MAX_DESC; i++ ) {
		RxDescRing[i].link = desc_base_phys_addr + ((i+1)%MAX_DESC)*sizeof(descriptor_t);
		RxDescRing[i].cmd = MAX_PACKET_SIZE;
		RxDescRing[i].ptr = buff_base_phys_addr +i*BUFFER_SIZE;
		RxDescRing[i].virt_next = &RxDescRing[(i+1)%MAX_DESC];
		RxDescRing[i].virt_buff = buff_base_virt_addr + i*BUFFER_SIZE;
	}

	TxDescRing = desc_base_virt_addr + MAX_DESC;
	for( i = 0; i < MAX_DESC; i++ ) {
		TxDescRing[i].link = desc_base_phys_addr + MAX_DESC*sizeof(descriptor_t)+ ((i+1)%MAX_DESC)*sizeof(descriptor_t);
		TxDescRing[i].cmd = MAX_PACKET_SIZE;
		TxDescRing[i].ptr = buff_base_phys_addr + ((i+MAX_DESC)*BUFFER_SIZE);
		TxDescRing[i].virt_next = &TxDescRing[(i+1)%MAX_DESC];
		TxDescRing[i].virt_buff = buff_base_virt_addr + ((i+MAX_DESC)*BUFFER_SIZE);
	}
	
	d->Rx.Curr = RxDescRing;
	d->Tx.Curr = TxDescRing;

	d->Rx.CurrInt = RxDescRing;
	d->Tx.CurrInt = TxDescRing;

		
	write32(REG_RXDP, desc_base_phys_addr);		/* set the initial rx descriptor	*/

	i = desc_base_phys_addr+MAX_DESC*sizeof(descriptor_t);
	write32(REG_TXDP, i);							/* set the initial tx descriptor	*/

	return B_OK;
}



uint16 reverse(uint16 val)
{
	int i = 0;
	uint16 retval = 0;
	uint16 mask = 1;
	
	dprintf("reverse in: %x\n", (unsigned int)val);
	for( i = 0; i < 16; i++ ) {		//	For every bit
		retval = retval<<1;			//	  shift the retval left once
		if( mask & val )			//	  if the current bit is set
			retval |= 1;			//	    set the 0 bit in retval
		mask = mask<<1;				//	  shift the mask left once
	}
	dprintf("reverse in: %x\n", (unsigned int)retval);
	
	return retval;
}

static void get_mac_address(dev_info_t *d)
{
	uint16 prev_eedata;
	int j;
	int i;
	uint16 prev;
	uint16 curr;
	uint16 *mac = (uint16*)d->mac_address.ebyte;
			
	prev = sis_eeprom_getword(d, 6);
	for (i = 0; i < 3; i++) {
		uint16 curr;
		uint16 temp;

		curr = sis_eeprom_getword(d, i+7);
		temp = (curr >> 1) | (prev << 15);
		prev = curr;

		temp = reverse(temp);
		mac[i] = temp;
	}
}


/* set hardware so all packets are received. */
static status_t setpromisc(dev_info_t * d)
{
	return(B_ERROR);
}


static status_t domulti(dev_info_t *d, uint8 *addr)
{
	dprintf("Set up multicast filter here\n");
	return (B_ERROR);
}


static void clear_statistics(dev_info_t *d)
{
	memset(&d->stats, 0, sizeof(stats_t));	
}



/*
 * Allocate and initialize semaphores and spinlocks.
 */
static status_t allocate_resources(dev_info_t *d)
{
	/* intialize rx semaphore with zero received packets		*/
	if ((d->Rx.Sem = create_sem(0, kDevName " rx")) < 0) {
		dprintf(kDevName " create rx sem failed %x \n", d->Rx.Sem);
		return (d->Rx.Sem);
	}
	set_sem_owner(d->Rx.Sem, B_SYSTEM_TEAM);

	
	/* intialize tx semaphore with the number of free tx buffers */
	if ((d->Tx.Sem = create_sem(MAX_DESC, kDevName " tx")) < 0) {
		delete_sem(d->Rx.Sem);
		dprintf(kDevName " create read sem failed %x \n", d->Tx.Sem);
		return (d->Tx.Sem);
	}
	
	set_sem_owner(d->Tx.Sem, B_SYSTEM_TEAM);

	d->blockFlag = 0;
	
	return (B_OK);
}





static void free_resources(dev_info_t *d)
{
		delete_sem(d->Rx.Sem);
		delete_sem(d->Tx.Sem);
}





#define SIS_EECTL		0x08

#define SIS_EECTL_DIN		0x00000001
#define SIS_EECTL_DOUT		0x00000002
#define SIS_EECTL_CLK		0x00000004
#define SIS_EECTL_CSEL		0x00000008

#define SIS_EECMD_WRITE		0x140
#define SIS_EECMD_READ		0x180
#define SIS_EECMD_ERASE		0x1c0


#define SIO_SET(x) {	register uint32 temp = read32(SIS_EECTL); temp |= x; write32(SIS_EECTL, temp); }

#define SIO_CLR(x) {	register uint32 temp = read32(SIS_EECTL); temp &= ~x; write32(SIS_EECTL, temp); }



static void sis_delay(dev_info_t *d)
{
	int			idx;

	for (idx = (300 / 33) + 1; idx > 0; idx--)
		read32(REG_CR);

	return;
}





static void sis_eeprom_idle(dev_info_t	*d)
{
	register int	i;

	SIO_SET(SIS_EECTL_CSEL);
	sis_delay(d);
	SIO_SET(SIS_EECTL_CLK);
	sis_delay(d);

	for (i = 0; i < 25; i++) {
		SIO_CLR(SIS_EECTL_CLK);
		sis_delay(d);
		SIO_SET(SIS_EECTL_CLK);
		sis_delay(d);
	}

	SIO_CLR(SIS_EECTL_CLK);
	sis_delay(d);
	SIO_CLR(SIS_EECTL_CSEL);
	sis_delay(d);
	write32(SIS_EECTL, 0x00000000);

	return;
}




/*
 * Send a read command and address to the EEPROM, check for ACK.
 */
static void
sis_eeprom_putbyte(d, addr)
	dev_info_t	*d;
	int			addr;
{
	register int		data, i;

	data = addr | SIS_EECMD_READ;

	/*
	 * Feed in each bit and stobe the clock.
	 */
	for (i = 0x400; i; i >>= 1) {
		if (data & i) {
			SIO_SET(SIS_EECTL_DIN);
		} else {
			SIO_CLR(SIS_EECTL_DIN);
		}
		sis_delay(d);
		SIO_SET(SIS_EECTL_CLK);
		sis_delay(d);
		SIO_CLR(SIS_EECTL_CLK);
		sis_delay(d);
	}
	
	return;
}





/*
 * Read a word of data stored in the EEPROM at address 'addr.'
 */
static uint16 sis_eeprom_getword(dev_info_t	*d, int addr)
{
	register int	i;
	uint16			word = 0;

	/* Force EEPROM to idle state. */
	sis_eeprom_idle(d);

	/* Enter EEPROM access mode. */
	sis_delay(d);
	SIO_CLR(SIS_EECTL_CLK);
	sis_delay(d);
	SIO_SET(SIS_EECTL_CSEL);
	sis_delay(d);

	/*
	 * Send address of word we want to read.
	 */
	sis_eeprom_putbyte(d, addr);

	/*
	 * Start reading bits from EEPROM.
	 */
	for (i = 0x8000; i; i >>= 1) {
		uint32 temp;
		SIO_SET(SIS_EECTL_CLK);
		sis_delay(d);
		temp = read32(SIS_EECTL);
		if( temp & SIS_EECTL_DOUT)
			word |= i;
		sis_delay(d);
		SIO_CLR(SIS_EECTL_CLK);
		sis_delay(d);
	}

	/* Turn off EEPROM access mode. */
	sis_eeprom_idle(d);

	return word;
}




/*
 * Read a sequence of words from the EEPROM.
 */
static void
sis_read_eeprom(d, dest, off, cnt, swap)
	dev_info_t	*d;
	caddr_t			dest;
	int			off;
	int			cnt;
	int			swap;
{
	int			i;
	uint16		word = 0, *ptr;

	for (i = 0; i < cnt; i++) {
		word = sis_eeprom_getword(d, off + i);
		ptr = (uint16*)(dest + (i * 2));
		if (swap)
			*ptr = ntohs(word);
		else
			*ptr = word;
	}

	return;
}




/*
static void sis_read_mac(dev_info_t *d, dev, dest)
	device_t		dev;
	caddr_t			dest;
{
	uint32		filtsave, csrsave;

	filtsave = read32(SIS_RXFILT_CTL);
	csrsave = read32(REG_CR);

	write32(REG_CR, REG_CR_RELOAD | filtsave);
	write32(REG_CR, 0);
		
	write32(SIS_RXFILT_CTL, filtsave & ~SIS_RXFILTCTL_ENABLE);

	write32(SIS_RXFILT_CTL, SIS_FILTADDR_PAR0);
	((uint16 *)dest)[0] = read16(SIS_RXFILT_DATA);
	write32(SIS_RXFILT_CTL,SIS_FILTADDR_PAR1);
	((uint16 *)dest)[1] = read16(SIS_RXFILT_DATA);
	write32(SIS_RXFILT_CTL, SIS_FILTADDR_PAR2);
	((uint16 *)dest)[2] = read16(SIS_RXFILT_DATA);

	write32(SIS_RXFILT_CTL, filtsave);
	write32(REG_CR, csrsave);
	return;
}
*/

int32	api_version = B_CUR_DRIVER_API_VERSION;
